"""Login to ``skore hub``."""

from __future__ import annotations

from contextlib import suppress
from datetime import datetime
from time import sleep
from webbrowser import open as open_webbrowser

from httpx import HTTPError, HTTPStatusError, TimeoutException
from rich.align import Align
from rich.panel import Panel

from .. import console
from .logout import logout
from .token import TokenError, access
from .token import persist as persist_token
from .uri import URI, URIError
from .uri import persist as persist_URI


def get_oauth_device_login(success_uri: str | None = None) -> tuple[str, str, str]:
    """
    Initiate device OAuth flow.

    Initiates the OAuth device flow.
    Provides the user with a URL and a OTP code to authenticate the device.

    Parameters
    ----------
    success_uri : str, optional
        The URI to redirect to after successful authentication.
        If not provided, defaults to None.

    Returns
    -------
    tuple
        A tuple containing:
        - authorization_url: str
            The URL to which the user needs to navigate
        - device_code: str
            The device code used for authentication
        - user_code: str
            The user code that needs to be entered on the authorization page
    """
    from skore_hub_project.client.client import HUBClient

    url = "identity/oauth/device/login"
    params = {"success_uri": success_uri} if success_uri is not None else {}

    with HUBClient(authenticated=False) as client:
        response = client.get(url, params=params).json()

        return (
            response["authorization_url"],
            response["device_code"],
            response["user_code"],
        )


def get_oauth_device_code_probe(device_code: str, *, timeout: int = 600) -> None:
    """
    Ensure authorization code is acknowledged.

    Start polling, wait for the authorization code to be acknowledged by the hub.
    This is mandatory to be authorize to exchange with a token.

    Parameters
    ----------
    device_code : str
        The device code to exchange for tokens.
    """
    from skore_hub_project.client.client import HUBClient

    url = "identity/oauth/device/code-probe"
    params = {"device_code": device_code}

    with HUBClient(authenticated=False) as client:
        start = datetime.now()

        while True:
            try:
                client.get(url, params=params)
            except HTTPStatusError as exc:
                if exc.response.status_code != 400:
                    raise

                if (datetime.now() - start).total_seconds() >= timeout:
                    raise TimeoutException("Authentication timeout") from exc

                sleep(0.5)
            else:
                break


def post_oauth_device_callback(state: str, user_code: str) -> None:
    """
    Validate the user-provided device code.

    This endpoint verifies the code entered by the user during the device auth flow.

    Parameters
    ----------
    state: str
        The unique value identifying the device flow.
    user_code: str
        The code entered by the user.
    """
    from skore_hub_project.client.client import HUBClient

    url = "identity/oauth/device/callback"
    data = {"state": state, "user_code": user_code}

    with HUBClient(authenticated=False) as client:
        client.post(url, data=data)


def get_oauth_device_token(device_code: str) -> tuple[str, str, str]:
    """
    Exchanges the device code for an access token.

    This endpoint completes the device authorization flow
    by exchanging the validated device code for an access token.

    Parameters
    ----------
    device_code : str
        The device code to exchange for tokens

    Returns
    -------
    tuple
        A tuple containing:
        - access_token : str
            The OAuth access token
        - refresh_token : str
            The OAuth refresh token
        - expires_at : str
            The expiration datetime as ISO 8601 str of the access token
    """
    from skore_hub_project.client.client import HUBClient

    url = "identity/oauth/device/token"
    params = {"device_code": device_code}

    with HUBClient(authenticated=False) as client:
        response = client.get(url, params=params).json()
        tokens = response["token"]

        return (
            tokens["access_token"],
            tokens["refresh_token"],
            tokens["expires_at"],
        )


def login(*, timeout: int = 600) -> None:
    """Login to ``skore hub``."""
    with suppress(URIError, TokenError, HTTPError):
        # Avoid re-login when
        # - the persisted URI is not conflicting with the envar
        # - the persisted token is valid or can be refreshed
        URI()
        access(refresh=True)

        console.print(
            Panel(
                Align(
                    "Already logged in!",
                    align="center",
                ),
                title="[cyan]Skore Hub[/cyan]",
                border_style="orange1",
                expand=False,
                padding=1,
                title_align="center",
            )
        )

        return

    logout()

    url, device_code, user_code = get_oauth_device_login()

    console.rule("[cyan]Skore Hub[/cyan]")
    console.print(
        f"Opening browser; if this fails, please visit this URL to log in:\n{url}",
        soft_wrap=True,
    )

    open_webbrowser(url)

    get_oauth_device_code_probe(device_code, timeout=timeout)
    post_oauth_device_callback(device_code, user_code)

    persist_token(*get_oauth_device_token(device_code))
    persist_URI(URI())



"""URI used for ``skore hub`` authentication."""

from os import environ
from pathlib import Path
from tempfile import gettempdir
from typing import Final

DEFAULT: Final[str] = "https://api.skore.probabl.ai"
ENVARNAME: Final[str] = "SKORE_HUB_URI"


class URIError(RuntimeError):
    """An exception dedicated to ``URI``."""


def Filepath() -> Path:
    """Filepath used to save the URI on disk."""
    return Path(gettempdir(), "skore.uri")


def persist(uri: str) -> None:
    """Persist the URI."""
    Filepath().write_text(uri)


def URI() -> str:
    """
    URI used for ``skore hub`` authentication.

    Notes
    -----
    It is discovered by following the rules in order:

    1. If you have setup an API key using the envar ``SKORE_HUB_API_KEY``, returns the
       value of the envar ``SKORE_HUB_URI``, or ``https://api.skore.probabl.ai``.

    2. If you have an active token, extract the content of the envar ``SKORE_HUB_URI``
       and the URI associated with the token:
        2.1. If both the URIs from the environment and the one associated with the token
             are empty, returns ``https://api.skore.probabl.ai``.
        2.2. If the URI from the environment is empty and the one associated with the
             token isn't, returns the URI associated with the token.
        2.3. If the URI associated with the token is empty, and the one from the
             environment isn't, returns the URI from the environment.
        2.4. If both the URIs associated with the token and the one from the
             environment are not empty:
                2.4.1. If both are equal, returns one of them.
                2.4.2. If both aren't equal, raises a conflicting exception.
    """
    uri_from_environment = environ.get(ENVARNAME)

    if environ.get("SKORE_HUB_API_KEY"):
        return uri_from_environment or DEFAULT

    filepath = Filepath()
    uri_from_persistence = filepath.read_text() if filepath.exists() else None

    if (
        uri_from_persistence
        and uri_from_environment
        and (uri_from_persistence != uri_from_environment)
    ):
        raise URIError(
            f"\nBad condition: the persisted URI is conflicting with the environment:\n"
            f"\tFrom {filepath}: '{uri_from_persistence}'\n"
            f"\tFrom ${ENVARNAME}: '{uri_from_environment}'\n"
            f"\nPlease run `skore-hub-logout`."
        )

    return uri_from_persistence or uri_from_environment or DEFAULT



"""Token used for ``skore hub`` authentication."""

from datetime import datetime, timezone
from json import dumps, loads
from pathlib import Path
from tempfile import gettempdir


class TokenError(Exception):
    """An exception dedicated to ``Token``."""


def post_oauth_refresh_token(refresh_token: str) -> tuple[str, str, str]:
    """
    Refresh an access token using a provided refresh token.

    This endpoint allows a client to obtain a new access token
    by providing a valid refresh token.

    Parameters
    ----------
    refresh_token : str
        A valid refresh token

    Returns
    -------
    tuple
        A tuple containing:
        - access_token : str
            The OAuth access token
        - refresh_token : str
            The OAuth refresh token
        - expires_at : str
            The expiration datetime as ISO 8601 str of the access token
    """
    from skore_hub_project.client.client import HUBClient

    url = "identity/oauth/token/refresh"
    json = {"refresh_token": refresh_token}

    with HUBClient(authenticated=False) as client:
        response = client.post(url, json=json).json()

        return (
            response["access_token"],
            response["refresh_token"],
            response["expires_at"],
        )


def Filepath() -> Path:
    """Filepath used to persist the token on disk."""
    return Path(gettempdir(), "skore.token")


def persist(access: str, refreshment: str, expiration: str) -> None:
    """
    Persist the token.

    Persist the token on disk to prevent user to login more than once, as long as
    the access token is valid or can be refreshed.
    """
    Filepath().write_text(
        dumps(
            [
                access,
                refreshment,
                expiration,
            ]
        )
    )


def access(*, refresh: bool = True) -> str:
    """
    Access token.

    Parameters
    ----------
    refresh : bool, optional
        Refresh the token on-the-fly if necessary, default True.
    """
    if not Filepath().exists():
        raise TokenError("You are not logged in. Please run `skore-hub-login`.")

    access: str
    refreshment: str
    expiration: str

    access, refreshment, expiration = loads(Filepath().read_text())

    if refresh and datetime.fromisoformat(expiration) <= datetime.now(timezone.utc):
        # Retrieve freshly updated tokens
        access, refreshment, expiration = post_oauth_refresh_token(refreshment)

        # Re-save the refreshed tokens
        persist(access, refreshment, expiration)

    return access



"""Definition of the payload used to associate feature importance media with report."""

from __future__ import annotations

from abc import ABC
from collections.abc import Callable
from functools import reduce
from typing import TYPE_CHECKING, ClassVar, Literal, cast

from orjson import OPT_NON_STR_KEYS, OPT_SERIALIZE_NUMPY, dumps

from skore_hub_project.artifact.media.media import Media, Report
from skore_hub_project.protocol import Display, EstimatorReport

if TYPE_CHECKING:
    from pandas import DataFrame


class FeatureImportance(Media[Report], ABC):  # noqa: D101
    accessor: ClassVar[str]
    content_type: Literal["application/vnd.dataframe"] = "application/vnd.dataframe"

    def content_to_upload(self) -> bytes | None:  # noqa: D102
        try:
            function = cast(
                "Callable[..., Display | DataFrame]",
                reduce(getattr, self.accessor.split("."), self.report),
            )
        except AttributeError:
            return None

        result = function()

        if isinstance(result, Display):
            result = result.frame()

        return dumps(
            result.fillna("NaN").to_dict(orient="tight"),
            option=(OPT_NON_STR_KEYS | OPT_SERIALIZE_NUMPY),
        )


class Permutation(FeatureImportance[EstimatorReport], ABC):  # noqa: D101
    accessor: ClassVar[str] = "feature_importance.permutation"
    name: Literal["permutation"] = "permutation"

    def content_to_upload(self) -> bytes | None:  # noqa: D102
        for key, obj in reversed(list(self.report._cache.items())):
            if len(key) < 7:
                continue

            if len(key) == 7:
                parent_hash, metric, data_source, scoring, *_ = key
            else:
                parent_hash, metric, data_source, data_source_hash, scoring, *_ = key

            if (
                parent_hash == self.report._hash
                and metric == "permutation_importance"
                and data_source == self.data_source
                and scoring is None
            ):
                return dumps(
                    obj.fillna("NaN").to_dict(orient="tight"),
                    option=(OPT_NON_STR_KEYS | OPT_SERIALIZE_NUMPY),
                )

        return None


class PermutationTrain(Permutation):  # noqa: D101
    data_source: Literal["train"] = "train"


class PermutationTest(Permutation):  # noqa: D101
    data_source: Literal["test"] = "test"


class MeanDecreaseImpurity(FeatureImportance[EstimatorReport]):  # noqa: D101
    accessor: ClassVar[str] = "feature_importance.mean_decrease_impurity"
    name: Literal["mean_decrease_impurity"] = "mean_decrease_impurity"
    data_source: None = None


class Coefficients(FeatureImportance[Report]):  # noqa: D101
    accessor: ClassVar[str] = "feature_importance.coefficients"
    name: Literal["coefficients"] = "coefficients"
    data_source: None = None



"""Logout from ``skore hub``."""

from . import token, uri


def logout() -> None:
    """Logout from ``skore hub`` by deleting the persisted token."""
    token.Filepath().unlink(missing_ok=True)
    uri.Filepath().unlink(missing_ok=True)
